;;;; Copyright (c) 2004-2005, Dirk H.P. Gerrits
;;;; All rights reserved.
;;;;
;;;; This software is licensed under the terms of the BSD license, stated in the
;;;; file named LICENSE.
;;;;
;;;;----------------------------------------------------------------------------
;;;;
;;;; This file provides a system manager for process linking.
;;;;
;;;;----------------------------------------------------------------------------

(in-package :erlisp)

(defvar *link-table* (make-hash-table))

(defun kill-process (target)
  (setf (slot-value target 'time-to-die) t))

(defun suicide (&optional (reason 'normal))
  (when (not (slot-value (current-process) 'dying))
    (setf (slot-value (current-process) 'dying) t)
    (error reason)))

(defun listen-to-system-manager ()
  (when (slot-value (current-process) 'time-to-die)
    (suicide "System manager requested termination.")))

(defun system-manager-process ()
  (do ((msg (receive-with-matcher (cond-matcher m) (t m)) 
	    (receive-with-matcher (cond-matcher m) (t m))))
      (nil)
    (cond ((eq 'link (first msg))
	   (setf (gethash (list (second msg) (third msg)) 
			  *link-table*)
		 (fourth msg)))
	  ((eq 'exit (first msg))
	   (maphash #'(lambda (k v)
			(let ((src (second msg))
			      (reason (third msg)))
			  (when (member src k)
			    (when (eq src (first k))
			      (cond ((or (eq v 'notify) (and (eq v 'default) (eq reason 'normal)))
				     (send (second k) msg))
				    (t (kill-process (second k)))))
			    (remhash k *link-table*))))
		    *link-table*)))))

(defvar *system-manager* (spawn #'system-manager-process))

(defun link (process &key (link-type 'default) (recipient (current-process)))
  (send  *system-manager*
	(list 'link process recipient link-type)))

(defun send-exit-message (&optional (reason 'normal))
  (setf (slot-value (current-process) 'dying) t)
  (send *system-manager*
	(list 'exit (current-process) reason)))
